#!/usr/bin/env python3
# coding=utf-8

# This file is part of PROMISE.
#
# 	PROMISE is free software: you can redistribute it and/or modify it
# 	under the terms of the GNU Lesser General Public License as
# 	published by the Free Software Foundation, either version 3 of the
# 	License, or (at your option) any later version.
#
# 	PROMISE is distributed in the hope that it will be useful, but WITHOUT
# 	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# 	or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# 	Public License for more details.
#
# 	You should have received a copy of the GNU Lesser General Public
# 	License along with PROMISE. If not, see
# 	<http://www.gnu.org/licenses/>.
#
# Promise v1 was written by Romain Picot
# Promise v2 has been written from v1 by Thibault Hilaire and Sara Hoseininasab
# 	Sorbonne Université
# 	LIP6 (Computing Science Laboratory)
# 	Paris, France
# Contact: thibault.hilaire@lip6.fr
#
#
# File: runPromise.py
# Date: April 2020
#
# 	contain the entry function, called to run Promise
#
# 	© Thibault HILAIRE, April 2020


__doc__= """
\U0001f918 Promise v2 \U0001f918

Usage:
	runPromise -h | --help
	runPromise (hsd|hs|sd) [options]

Options:
  -h --help                     Show this screen.
  --conf CONF_FILE              get the configuration file [default: promise.yml]
  --output OUTPUT               set the path of the output (where the result files are put)
  --verbosity VERBOSITY         set the verbosity (betwen 0  and 4 for very low level debug) [default: 1]
  --log LOGFILE                 set the log file (no log file if this is not defined)
  --verbosityLog VERBOSITY      set the verbosity of the log file
  --debug                       put intermediate files into `debug/` (and `compileErrors/` for compilation errrors) and display the execution trace when an error comes
  --run RUN                     file to be run
  --compile COMMAND             command to compile the code
  --files FILES                 list of files to be examined by Promise (by default, all the .cc files)
  --nbDigits DIGITS             general required number of digits
  --path PATH                   set the path of the project (by default, the current path)
  --pause                       do pause between steps
  --noParsing                   do not parse the C file (__PROMISE__ are replaced and that's all)
  --alias ALIAS                 allow aliases (examples "g++=g++-14") [default:""]
  hsd                           Half/Single/Double mixed-precision
  hs                            Half/Single mixed-precision
  sd                            Single/Double mixed-precision

  
"""

from os.path import join
import sys
from collections import Counter

from .promise import Promise
from .utils import parseOptions, PromiseError, getYMLOptions, Timing, pause, commaAnd
from .utils import customArgParser, __params__
from .logger import PrLogger
from .prfile import PrFile


def runPromise(argv=None):
	"""This function is registered (in setup.py) as an entry_point
	argv is used for the unit tests"""

	# reset the logger and get a new instance
	logger = PrLogger()
	displayTrace = False


	# types handle by Promise
	typeNames = {'h': 'Half', 's': 'Single', 'd': 'Double', 'q': 'Quad'}
	types = {'h': 'half_float::half', 's': 'float', 'd': 'double', 'q': 'float128'}


	try:
		# import docopt
		# get the options from command line and yml file
		# print("**** argv:", argv)
		args = customArgParser(sys.argv[1:] if argv is None else argv, __params__)      # parse the command line
		
		# args = docopt.docopt(__doc__, argv=sys.argv[1:] if argv is None else argv)      # parse the command line
		# print("**** args: ", args)
		displayTrace = args['--debug']
		logger.configureLogger(args)                                                   # configure the logger
		options = getYMLOptions(args)                                           # get the options from the yml file
		logger.message("\U0001f918 Promise v2 \U0001f918")
		method, path, files, run, nbDigits, compileLines, outputPath, typeCustom, alias = parseOptions(options)    # parse the options
		PrFile.setCustom(typeCustom)
		compileErrorPath = join(path, 'compileErrors') if args['--debug'] else None
		tempPath = join(path, 'debug') if args['--debug'] else None


		
		# run with timing
		with Timing() as timing:
			# create Promise object
			pr = Promise(path, files, run, nbDigits, compileLines, parsing=not args['--noParsing'], alias=alias)

			# display general infos
			logger.info("We are working with %d file%s and %d different types" %
			            (pr.nbFiles, ('' if pr.nbFiles < 2 else 's'), pr.nbVariables))
			logger.info(pr.expectation())

			# debug the files
			if args['--debug']:
				pr.exportParsedFiles(tempPath)


			print("**** test - tempPath ", tempPath)
			# get the cadna reference
			highest = types['q'] if 'q' in method else types['d']
			pr.changeSameType(highest)
			logger.step("Get a reference result with Cadna (%s)" % highest)
			pr.compileAndRun(tempPath, cadna=True)
			if args['--pause']:
				pause()

			# try with the highest precision
			logger.step("Check with highest format (%s)" % typeNames[method[-1]])
			pr.changeSameType(types[method[-1]])
			
			if not pr.compileAndRun(tempPath):
				pr.changeSameType(highest)
				raise PromiseError("You should lower your expectation, it doesn't work with " + typeNames[method[-1]])
			
			if args['--pause']:
				pause()

			# do the Delta-Debug passes ('s','d' and then 'h','s' when method is 'hsd' for example)
			for lo, hi in reversed(list(zip(method, method[1:]))):
				logger.step("Delta-Debug %s/%s" % (typeNames[lo], typeNames[hi]))
				res = pr.runDeltaDebug(types[lo], types[hi], tempPath, args['--pause'], compileErrorPath)
				# stop if the DeltaDebug is not successful
				if not res:
					break

		# export the output
		if argv is None:
			pr.exportFinalResult(outputPath)


	except PromiseError as e:
		logger.error(e, exc_info=displayTrace)

	else:
		if timing:
			# display the number of each type
			count = Counter(pr.typesDict.values())  # count the nb of each type (result is a dictionary type:nb)
			li = ["%dx %s" % (v, k) for k, v in count.items()]
			logger.message("The final result contains %s.", commaAnd(li))
			logger.debug("Final types:\n" + pr.strResult())

			# display the stats
			logger.message("It tooks %.2fs", timing.timing)
			logger.message("\U0001F449 %d compilations (%d failed) for %.2fs", *pr.compilations)
			logger.message("\U0001F449 %d executions   (%d failed) for %.2fs", *pr.executions)

			logger.reset()
			return pr.setPerType()

	# reset the handlers, in case of running runPromise several times (otherwise, the log files are still open)
	logger.reset()


def runPromise_custom(argv=None):
	"""This function is registered (in setup.py) as an entry_point
	argv is used for the unit tests"""

	# reset the logger and get a new instance
	logger = PrLogger()
	displayTrace = False


	# types handle by Promise
	typeNames = {'h': 'Half', 's': 'Single', 'd': 'Double', 'q': 'Quad'}
	types = {'h': 'half_float::half', 's': 'float', 'd': 'double', 'q': 'float128'}


	try:
		# get the options from command line and yml file
		args = customArgParser(sys.argv[1:] if argv is None else argv, __params__)      # parse the command line
		displayTrace = args['--debug']
		logger.configureLogger(args)                                                   # configure the logger
		options = getYMLOptions(args)                                           # get the options from the yml file
		logger.message("\U0001f918 Promise v2 \U0001f918")
		method, path, files, run, nbDigits, compileLines, outputPath, typeCustom, alias = parseOptions(options)    # parse the options
		PrFile.setCustom(typeCustom)
		compileErrorPath = join(path, 'compileErrors') if args['--debug'] else None
		tempPath = join(path, 'debug') if args['--debug'] else None

		# run with timing
		with Timing() as timing:
			# create Promise object
			pr = Promise(path, files, run, nbDigits, compileLines, parsing=not args['--noParsing'], alias=alias)

			# display general infos
			logger.info("We are working with %d file%s and %d different types" %
			            (pr.nbFiles, ('' if pr.nbFiles < 2 else 's'), pr.nbVariables))
			logger.info(pr.expectation())

			# debug the files
			if args['--debug']:
				pr.exportParsedFiles(tempPath)

			# get the cadna reference
			highest = types['q'] if 'q' in method else types['d']
			pr.changeSameType(highest)
			logger.step("Get a reference result with Cadna (%s)" % highest)
			pr.compileAndRun(tempPath, cadna=True)
			
			if args['--pause']:
				pause()

			# try with the highest precision
			logger.step("Check with highest format (%s)" % typeNames[method[-1]])
			pr.changeSameType(types[method[-1]])
			
			if not pr.compileAndRun(tempPath):
				pr.changeSameType(highest)
				raise PromiseError("You should lower your expectation, it doesn't work with " + typeNames[method[-1]])
			
			if args['--pause']:
				pause()

			# do the Delta-Debug passes ('s','d' and then 'h','s' when method is 'hsd' for example)
			for lo, hi in reversed(list(zip(method, method[1:]))):
				logger.step("Delta-Debug %s/%s" % (typeNames[lo], typeNames[hi]))
				res = pr.runDeltaDebug(types[lo], types[hi], tempPath, args['--pause'], compileErrorPath)
				# stop if the DeltaDebug is not successful
				if not res:
					break

		# export the output
		if argv is None:
			pr.exportFinalResult(outputPath)


	except PromiseError as e:
		logger.error(e, exc_info=displayTrace)


	else:
		if timing:
			# display the number of each type
			count = Counter(pr.typesDict.values())  # count the nb of each type (result is a dictionary type:nb)
			li = ["%dx %s" % (v, k) for k, v in count.items()]
			logger.message("The final result contains %s.", commaAnd(li))
			logger.debug("Final types:\n" + pr.strResult())

			# display the stats
			logger.message("It tooks %.2fs", timing.timing)
			logger.message("\U0001F449 %d compilations (%d failed) for %.2fs", *pr.compilations)
			logger.message("\U0001F449 %d executions   (%d failed) for %.2fs", *pr.executions)

			logger.reset()
			return pr.setPerType()

	# reset the handlers, in case of running runPromise several times (otherwise, the log files are still open)
	logger.reset()




def runPromise_custom_test(argv=None):
	"""This function is registered (in setup.py) as an entry_point
	argv is used for the unit tests"""

	print("**** start")
	# reset the logger and get a new instance
	logger = PrLogger()
	displayTrace = False

	print("****1. argv:", argv)
	
	# types handle by Promise
	typeNames = {'h': 'Half', 's': 'Single', 'd': 'Double', 'q': 'Quad'}
	types = {'h': 'half_float::half', 's': 'float', 'd': 'double', 'q': 'float128'}

	print("*****sys.argv:", sys.argv)
	try:
		# get the options from command line and yml file
		args = customArgParser(sys.argv[1:] if argv is None else argv, __params__)      # parse the command line
		# print("****2. args:", args)

		displayTrace = args['--debug']
		# print("****3. displayTrace:", displayTrace)

		logger.configureLogger(args)                                                   # configure the logger
		options = getYMLOptions(args)                                           # get the options from the yml file

		# print("****4. options:", options)
		logger.message(r"\U0001f918 Promise v2 \U0001f918")
		method, path, files, run, nbDigits, compileLines, outputPath, typeCustom, alias = parseOptions(options)    # parse the options

		# print("****5. method, path, files, run, nbDigits, compileLines, outputPath, typeCustom, alias:")
		# print(method, path, files, run, nbDigits, compileLines, outputPath, typeCustom, alias)

		PrFile.setCustom(typeCustom)
		compileErrorPath = join(path, 'compileErrors') if args['--debug'] else None
		tempPath = join(path, 'debug') if args['--debug'] else None


		
		# run with timing
		with Timing() as timing:
			# create Promise object
			pr = Promise(path, files, run, nbDigits, compileLines, parsing=not args['--noParsing'], alias=alias)

			# display general infos
			logger.info("We are working with %d file%s and %d different types" %
			            (pr.nbFiles, ('' if pr.nbFiles < 2 else 's'), pr.nbVariables))
			
			logger.info(pr.expectation())

			# debug the files
			if args['--debug']:
				pr.exportParsedFiles(tempPath)


			# print("**** test - tempPath ", tempPath)
			# get the cadna reference
			highest = types['q'] if 'q' in method else types['d']

			# print("***5. highest:", highest)
			pr.changeSameType(highest)
			logger.step("Get a reference result with Cadna (%s)" % highest)

			pr.compileAndRun(tempPath, cadna=True)
			if args['--pause']:
				pause()

			# try with the highest precision
			logger.step("Check with highest format (%s)" % typeNames[method[-1]])
			pr.changeSameType(types[method[-1]])
			
			if not pr.compileAndRun(tempPath):
				pr.changeSameType(highest)
				raise PromiseError("You should lower your expectation, it doesn't work with " + typeNames[method[-1]])
			
			if args['--pause']:
				pause()

			# do the Delta-Debug passes ('s','d' and then 'h','s' when method is 'hsd' for example)
			for lo, hi in reversed(list(zip(method, method[1:]))):
				logger.step("Delta-Debug %s/%s" % (typeNames[lo], typeNames[hi]))
				res = pr.runDeltaDebug(types[lo], types[hi], tempPath, args['--pause'], compileErrorPath)
				# stop if the DeltaDebug is not successful
				if not res:
					break

		# export the output
		if argv is None:
			pr.exportFinalResult(outputPath)


	except PromiseError as e:
		logger.error(e, exc_info=displayTrace)


	else:
		if timing:
			# display the number of each type
			count = Counter(pr.typesDict.values())  # count the nb of each type (result is a dictionary type:nb)
			li = ["%dx %s" % (v, k) for k, v in count.items()]
			logger.message(r"The final result contains %s.", commaAnd(li))
			logger.debug(r"Final types:\n" + pr.strResult())

			# display the stats
			logger.message(r"It tooks %.2fs", timing.timing)
			logger.message(r"\U0001F449 %d compilations (%d failed) for %.2fs", *pr.compilations)
			logger.message(r"\U0001F449 %d executions   (%d failed) for %.2fs", *pr.executions)

			logger.reset()
			return pr.setPerType()

	# reset the handlers, in case of running runPromise several times (otherwise, the log files are still open)
	logger.reset()




if __name__ == "__main__":
	runPromise()






# coding=utf-8

"""This file is part of PROMISE.

	PROMISE is free software: you can redistribute it and/or modify it
	under the terms of the GNU Lesser General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	PROMISE is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
	Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with PROMISE. If not, see
	<http://www.gnu.org/licenses/>.

Promise v1 was written by Romain Picot
Promise v2 has been written from v1 by Thibault Hilaire and Sara Hoseininasab
	Sorbonne Université
	LIP6 (Computing Science Laboratory)
	Paris, France
Contact: thibault.hilaire@lip6.fr


File: PrFile.py
Date: April 2020

	Some useful functions to parse the C/C++ files and replace the __PROMISE__ and __PR_xxx__ by some types.
	To do that, we have to parse the file, in order to know if the __PROMISE__ are in a comment, string, etc. or
	if it is used for a variable declaration, an array, a pointer, a function returns, a function argument, etc.
	Depending on the case, the code may be a little bit changed
	(`__PROMISE__ x,y=5;` is changed in `__PROMISE__ x; __PROMISE__ y; y=5;` for example)

	© Thibault HILAIRE, April 2020
"""

from .logger import PrLogger
from .errors import PromiseError
import regex as re
from itertools import zip_longest
from os.path import join, split, relpath

# logger
logger = PrLogger()

# WARNING: these regex (and the parsing done with them) is not robust to all the code
# if we want to be able to deal with these kind of cases (initialization part that is not just a litteral)
# we need to use a dedicated C parser (not home-made parser and few regex)

# regex to find the __PROMISE__, __PR_xxxx__ or __PRC_xxxx__
re_tag = re.compile('(.*?)(__PROMISE__|__PRC{0,1}_([^_]+)__)(.*)', re.DOTALL)

# regex for a variable
re_decl = re.compile(r'\s*(\**)\s*(\w+)((?:\[\w+\])*)\s*(.*)', re.DOTALL)
re_var_fun = re.compile(r'^(\**)\s*(\w+)(\[\w+\])*\s*(?:=\s*[\w\.]+)?')

# regex to remove the PROMISE_CHECK_VAR(...); and PROMISE_CHECK_ARRAY(...);
re_check = re.compile(r'PROMISE_CHECK_VAR\([\w]+\);|PROMISE_CHECK_ARRAY\([\w]+\);')


class VariableType:
	def __init__(self, key):
		self._key = key

	@property
	def value(self):
		return self._key


class CustomType(VariableType):
	pass


class NormalType(VariableType):
	pass


class Status:
	"""Enumeration for the status of a line
	(inside a /* .... */ comment, inside a //.... comment,
	inside a string "....." or not)"""
	NORMAL = 0
	COMMENT_C = 1
	COMMENT_CPP = 2
	STRING = 3
	PARENTHESIS = 4

	def __init__(self, status=None):
		"""create a empty stack for the status"""
		if status:
			self._stack = list(status.stack)
		else:
			self._stack = []

	def __eq__(self, other):
		"""compare to the last status"""
		if self._stack:
			return self._stack[-1] == other
		else:
			return Status.NORMAL == other

	def __ne__(self, other):
		"""negative compare"""
		return not (self == other)

	def __str__(self):
		d = {0: "NORMAL", 1: "COMMENT /*", 2: "COMMENT //", 3: "STRING", 4: "PARENTHESIS"}
		if not self._stack:
			return "NORMAL"
		else:
			return "[" + ", ".join(d[s] for s in self._stack) + "]"

	@property
	def stack(self):
		"""Returns the stack"""
		return self._stack

	def parseLine(self, line):
		"""Parse the line character by character to evalute the
		status at the line (see Status class), from the status before the line
		This is used to handle multi-line strings or multi-line comments"""
		# nothing to do for an empty line, otherwise, we parse it, char by char
		if line:
			prev = line[0]
			self._parseChar('', prev)
			for succ in line[1:]:
				# parse the two char to update the status stack
				change = self._parseChar(prev, succ)
				# if we have change the status, then c1 has been "consumed"
				# and should not be use for a new pattern (like in "/*/")
				prev = '' if change else succ


	def _parseChar(self, prev, cur):
		"""update the status stack according the current char (cur) and the previous char (prev)"""
		stsize = len(self._stack)
		# according to the status, check if we change the status
		if self == Status.NORMAL or self == Status.PARENTHESIS:
			# NORMAL status
			if prev == '/' and cur == '*':
				self._stack.append(Status.COMMENT_C)
			elif prev == '/' and cur == '/':
				self._stack.append(Status.COMMENT_CPP)
			elif cur == "\"":
				self._stack.append(Status.STRING)
			elif cur == "(":
				self._stack.append(Status.PARENTHESIS)
			# PARENTHESIS
			elif self == Status.PARENTHESIS and cur == ")":
				self._stack.pop()

		# are we inside a CPP comment // ?
		elif self == Status.COMMENT_CPP:
			if cur == '\n':
				self._stack.pop()
		# are we inside a C comment /* ?
		elif self == Status.COMMENT_C:
			if prev == '*' and cur == '/':
				self._stack.pop()
		# are we inside a string ?
		elif self == Status.STRING:
			if cur == "\"" and prev != "\\":
				self._stack.pop()

		# return True if something has changed
		return len(self._stack) != stsize


	def parseDecl(self, line):
		"""Generator that parses the declaration line
		(like 'x,y[24],*z, a = 25, b= toto(12,5);')
		returns (for each) the name of the variable, the code generated
		and then return the end of the line"""
		initPhase = False
		prev = line[0]
		deb = 0
		pos = 0
		for pos, cur in enumerate(line[1:], 2):
			if cur in [',', ')', ';'] and self == Status.NORMAL:
				# get the name of the variable, (pointer, array and init parts)
				decl = line[deb:pos-1]
				po, name, array, init = re_decl.match(decl).groups()
				# in case of initialization, separe declaration and initialization
				if initPhase:
					if array:
						decl = po + name + array + init + ";"
					else:
						decl = po + name + "; " + name + init + ";"
				else:
					decl += ';' if cur != ')' else ')'
				yield name, decl
				initPhase = False
				deb = pos
			# stop if we have ';' or '(' (not in initialization phase, ie *before* '=')
			if (cur == ';' or cur == ')') and self == Status.NORMAL:
				break
			elif cur == '(' and not initPhase and self == Status.NORMAL:
				# in that case, it is not a variable, but a function declaration
				decl = line[deb:pos]
				po, name, array, init = re_decl.match(decl).groups()
				yield name, decl
				break
			# if we have an '=', then we are after the initialization
			elif cur == '=' and self == Status.NORMAL:
				initPhase = True
			# parse the two char to update the status stack
			change = self._parseChar(prev, cur)
			prev = '' if change else cur
		else:
			if deb == 0:
				# it means that we have yield anything but we are at the end
				# (some special cases like __PROMISE__ x = (__PR__)xxx;)
				po, name, array, init = re_decl.match(line).groups()
				yield name, line

		# finally return the end of the line
		yield line[pos:]



class PrFile:
	"""PrFile class to handle the files
	a PrFile object correspond to a file that is treated by Promise
	the file is read, parsed and split at the __PROMISE__, __PR_xxx__ or __PRC_xxx__
	each of them correspond to a type that is handle by Promise (so transformed in quad, double, single or half)
	In a file, we keep a list of text and a list of indexes of C types
	To reconstruct the file, we just take the first part of text (first element in the list of text), then add the
	type corresponding to the first index , then the 2nd part of text, the 2nd types, etc.
	The types are taken from the dictionary of types (see the Promise object)
	"""

	_custom = {}    # a dictionary of custom types (used with __PRC_xxx__)

	def __init__(self, fileName, Pr, path, doParsing=True):
		"""Build the PrFile from the filename
		a Promise object is necessary (to register each __PROMISE__ or __PR_xxx__ type)"""
		self._fileName = fileName
		self._ltext = []        # list of strings
		self._typeKeys = []     # list of index of types
		self._path = path
		self._Pr = Pr

		# read and parse the file (split the file)
		try:
			with open(join(path, fileName), 'r') as file:
				if doParsing:
					self._complexConstruction(file, fileName)
				else:
					self._simpleConstruction(file)
		except FileNotFoundError:
			logger.error("the file %s doesn't exist !!", fileName)
			raise PromiseError("the file %s doesn't exist !!", fileName)


	@property
	def fileName(self):
		"""returns the fileName"""
		return self._fileName

	@classmethod
	def setCustom(cls, custom):
		"""Used to set the custom types
		(those used in __PRC_xxx__)"""
		cls._custom = custom

	def addKey(self, typeName, custom=False):
		"""add the typeName in the _typeKey list
		"""
		if custom:
			self._typeKeys.append(CustomType(typeName))
		else:
			self._typeKeys.append(NormalType(typeName))

	def treatVariableDeclaration(self, status, line, lineNb):
		"""this line is preceded by __PROMISE__ in the code
		this function replace a declaration like `__PROMISE x,y,z;` into multiple declarations
		like `__PROMISE__ x; __PROMISE__ y; __PROMISE__ z;`
		and treate them (add them to typeKeyps and ltext)
		Parameters
			-line: (string) line to consider
			- lineNb: (int) line number
		"""
		# function arguments or in a cast ?
		if self._ltext and status == Status.PARENTHESIS:
			# check if there is a variable within
			var_fun = re_var_fun.match(line)
			if var_fun:
				# register the variable
				self._Pr.registerVariable(var_fun.group(2), self._typeKeys[-1].value, self._fileName, lineNb)

		# a variable (or function) declaration
		elif self._ltext:
			# decompose the declarations/initializations (Ex 'x=12, y, z=5; blablabla')
			st = Status()
			res = list(st.parseDecl(line))
			variables = res[:-1]
			line = (' ' if isinstance(self._typeKeys[-1], NormalType) else '') + variables[-1][1].lstrip() + res[-1]
			for i, (name, varDecl) in enumerate(variables):
				# register the variable
				self._Pr.registerVariable(name, self._typeKeys[-1].value, self._fileName, lineNb)
				# add the type to text and typeKeys if it's not the  last one
				if i != len(variables)-1:
					self._ltext.append(' ' + varDecl.lstrip())
					# we use the same type as previously if it was a dedicated type (`__PR_xxx__`) and not `__PROMISE__`
					typeVar = self._Pr.registerType(None) if isinstance(self._typeKeys[-1].value, int) else self._typeKeys[-1].value
					self.addKey(typeVar, custom=isinstance(self._typeKeys[-1], CustomType))

		# at the end, we store the line
		self._ltext.append(line)

	def _simpleConstruction(self, file):
		"""Read the file and cut at __PROMISE__ or __PR_xxx__
		Old way to do, that do not consider strings, comment, mutliple variable declaration,
		declaration-and-initialization, etc."""
		endOfLine = ''
		for line in file:
			endOfLine = endOfLine + line
			mat = re_tag.match(endOfLine)
			while mat:
				self._ltext.append(mat.group(1))  # store the beginning of the line
				typeName = self._Pr.registerType(mat.group(3))  # get the key of the __PROMISE__ or __PR_xxx__ type
				self.addKey(typeName, mat.group(2).startswith('__PRC'))  # store the key
				endOfLine = mat.group(4)  # keep the end of the line
				mat = re_tag.match(endOfLine)  # match for the it
		self._ltext.append(endOfLine)

	def _complexConstruction(self, file, fileName):
		"""Read the file and do the parsing
		Srings, comments, mutliple variable declaration, declaration-and-initialization, etc.
		are considered"""
		lineNb = 0    # line number during the parsing
		status = Status()
		stack = []  # list of lines already parsed, and not yet added in self._ltext
		prevStatus = Status(status)
		for lineNb, line in enumerate(file):
			mat = re_tag.match(line)  # check for a __PROMISE__, __PR_xxx__ or __PRC_xxx__ inside line
			while mat:
				# parse the left part of the line and get the status
				status.parseLine(mat.group(1))
				if status == Status.NORMAL or status == Status.PARENTHESIS:
					if not mat.group(2).startswith('__PRC_'):
						# treat the 1st part (and the stack)
						self.treatVariableDeclaration(prevStatus, "".join(stack) + mat.group(1), lineNb)
					else:
						self._ltext.append("".join(stack) + mat.group(1))  # store the beginning of the line
					stack = []
					# get the key of the __PROMISE__ or __PR_xxx__ type
					typeName = self._Pr.registerType(mat.group(3))
					# store the key
					self.addKey(typeName, custom=mat.group(2).startswith('__PRC_'))
					prevStatus = Status(status)
				else:
					# we are in a comment or string
					# add the first part of the line and the __PROMISE__ in the stack
					stack.append(mat.group(1) + mat.group(2))
				# we continue on the rest of the line
				line = mat.group(4)
				mat = re_tag.match(line)
			# parse the line and add it the the stack
			status.parseLine(line)
			stack.append(line)
		# deal with the end of the file
		self.treatVariableDeclaration(prevStatus, "".join(stack), lineNb)

		# check the status at the end
		if status != Status.NORMAL:
			raise PromiseError("Something went wrong when parsing the file %s (status = %s)" % (fileName, status))


	def createFile(self, typeDict, path, prefix='', final=False, promiseHeader=False):
		"""create the file, where the __PROMISE__ and __PR_xxx__ are replaced by the vlaues in the `type` dictionary"""
		beginName, endName = split(self._fileName)
		filename = join(path, beginName, prefix + endName)
		logger.lowdebug("Create the file %s", filename)
		with open(filename, "w") as f:
			# include promise headers (except for the final version)
			if not final:
				f.write('#include <half.hpp>\n')
				relative = relpath(path='.', start=beginName)   # build the relative path from the file to the top folder (where promise.h is)
				incl = join(relative, "promise")
				f.write('#include "' + incl + ('_header.h"\n' if promiseHeader else '.h"\n'))
			for text, key in zip_longest(self._ltext, self._typeKeys):
				if final:
					f.write(re.sub(re_check, '', text))
				else:
					f.write(text)
				if key is not None:
					if isinstance(key, NormalType):
						f.write(typeDict[key.value])
					elif isinstance(key, CustomType):
						f.write(self._custom.get(typeDict[key.value], typeDict[key.value]))   # The CustomDict is skiped if the typeDict[] is not in the dictionary





# coding=utf-8

"""This file is part of PROMISE.

	PROMISE is free software: you can redistribute it and/or modify it
	under the terms of the GNU Lesser General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	PROMISE is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
	Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with PROMISE. If not, see
	<http://www.gnu.org/licenses/>.

Promise v1 was written by Romain Picot
Promise v2 has been written from v1 by Thibault Hilaire and Sara Hoseininasab
	Sorbonne Université
	LIP6 (Computing Science Laboratory)
	Paris, France
Contact: thibault.hilaire@lip6.fr


File: Promise.py
Date: April 2020

	Main class of Promise v2
	It deals with the files, is able to change the types, compile and run the program, in order to get the
	precision expectation.

	© Thibault HILAIRE, April 2020
"""


from tempfile import mkdtemp
from os.path import join
from os import getcwd, environ#, chmod
from stat import S_IXUSR
from pkg_resources import resource_filename
import re
from tqdm import tqdm
from math import ceil, log2, isnan, isinf
from colorama import Fore
from datetime import datetime

from .errors import PromiseError, PromiseCompilationError
from .utils import runCommand, cd, Timing, pause, commaAnd
from .prfile import PrFile
from .deltadebug.promisedd import PromiseDD

from .logger import PrLogger
from os.path import join, split

logger = PrLogger()


regDumpST = re.compile(
	r"^\[PROMISE_DUMP_ST\] (\w+)(?:\[(\d+)\])? = \(([\w\-+.]+),([\w\-+.]+),([\w\-+.]+)\), nb significant digits=(\d+)")
regDump = re.compile(r"^\[PROMISE_DUMP\] (\w+)(?:\[(\d+)\])? = ([\w\-+.]+)")


class Promise:
	"""
	principal class for Promise v2
	It mainly contains a list of files (list of PrFile objects) and a dictionary, that maps for every variable type
	considered by Promise (all the types in form __PROMISE__ or __PR_xxx__ in the C code) to a given type
	"""

	def __init__(self, path, files, run, nbDigits, compileLines, parsing=True, alias={}):
		"""Build a Promise object
		Parameters:
			- path: (strin) path where are the files
			- files: (list) filenames
			- run: (string) filename of the executable
			- nbDigits: (tuple (int,{string:int})) number of correct digits required and then nb of digits per variable
			- compileLines: (list of string) command lines to execute to create the `run` file
			- parsing: (boolean) True if we cleverly parse the C files
		"""
		self._path = path
		self._run = run
		self._nbDigits = nbDigits
		self._compile = compileLines
		self._types = {}            # key: 0, 1, 2 for __PROMISE__ defined types and 'xxx' for __PR_xxx__ types
		self._counter = 0           # index of the *next* __PROMISE__ type
		self._ref = {}              # dictionary of references (variable name: mean value or list of mean values)
		self._power = {}            # dictionary of 10^-nbDigits
		self._variables = {}        # dictionary of variables {typeName: (variableName, fileName, lineNumber)}
		self._compilations = [0, 0, 0]
		self._executions = [0, 0, 0]
		self._cache = {}
		self._alias = alias
		# parse each files
		self._files = [PrFile(f, self, path, doParsing=parsing) for f in files]

	@property
	def types(self):
		"""Returns the list of types (keys)"""
		return list(self._types.keys())

	@property
	def variables(self):
		"""Returns the list of variables"""
		return self._variables.values()

	@property
	def typesDict(self):
		"""Returns the dictionary of types"""
		return self._types

	def setPerType(self):
		""""Returns a dictionary that, for each type, give the set of variables of that type
		so if _types is {0: 'double', 1: 'float', 2, 'double', 'x': 'float', 3: 'quad'}, setPerType will return
		{'double': {0, 2}, 'float': {1, 'x'}, 'quad': {3}}"""
		d = {}
		for k, v in self._types.items():
			if v in d:
				d[v] |= {k}
			else:
				d[v] = {k}
		return d

	def strResult(self):
		"""Returns the result (a string with each variable and its name)"""
		d = {}
		for k, v in self._types.items():
			var = ("%s (%s:%d)" % self._variables[k]) if k in self._variables else ("type " + str(k))
			if v in d:
				d[v] |= {var}
			else:
				d[v] = {var}
		return "\n".join("%s:\n%s\n" % (t, commaAnd(list(l))) for t, l in d.items())

	def registerVariable(self, name, typeName, fileName, lineNb):
		"""Register a variable with its type, filename and line number"""
		# if several variables share the same type, only the last one is now considered
		# (not a problem, this is only used for nice display...)
		self._variables[typeName] = (name, fileName, lineNb)

	@property
	def compilations(self):
		"""Returns the nb of compilations and time of computations"""
		return self._compilations

	@property
	def executions(self):
		"""Returns the nb of executions and time of executions"""
		return self._executions

	@property
	def nbFiles(self):
		"""Returns the nb of files"""
		return len(self._files)

	@property
	def nbVariables(self):
		"""Returns the nb of variables"""
		return len(self._types)

	def expectation(self):
		"""Returns a string defining the expectation (nb of Digits)"""
		exp = "The expectation is %s digits" % self._nbDigits[0]
		if self._nbDigits[1]:
			exp += ' (and ' + commaAnd(['%d digits for variable `%s`' % (d, v) for v, d in self._nbDigits[1].items()]) + ')'
		return exp + '.'

	def registerType(self, typeName):
		"""when a __PROMISE__ or __PR_xxx__ type is found, we register it here
		(add its name in the dictionary _types) """
		if typeName:
			self._types[typeName] = None
			return typeName
		else:
			self._types[self._counter] = None
			self._counter += 1
			return self._counter - 1


	def changeSameType(self, Ctype):
		"""change the type for all the Types"""
		for k in self._types:
			self._types[k] = Ctype

	def changeSomeTypes(self, Ctype, lkeys):
		"""change the type of some keys"""
		for k in lkeys:
			self._types[k] = Ctype

	def changeTypes(self, dtype):
		"""change the type dictionary to a new dictionary"""
		for k in dtype:
			self._types[k] = dtype[k]

	def getTypesEqualTo(self, typ):
		"""returns the list of types (keys) equal to typ"""
		return [k for k, t in self._types.items() if t == typ]


	def compileAndRun(self, dest=None, cadna=False, result=True, compilationErrorAsDebug=False, compileErrorPath=None):
		"""
		- copy the project,
		- create the new files (from the original ones) if `createFiles` if True
		- compile it
		- run it
		- and get the result

		Parameters:
			- cadna: is True if Cadna is used (to create the reference)
			- result: True when we want the result (False otherwise)
			- compilationErrorasDebug: True when we just want to display the compilation error as debug
			- compileErrorCopy: (str) if not None, it copies the files in it if the compilation fails

		Returns True if the result is valid (with the nb of siginifcant digits greater thar required)"""

		logger.debug('runAndCompile with types=%s' % str(self._types))
		varFailed = None  # name of the variable that make the test failed

		# memoization
		if frozenset(self._types.items()) not in self._cache:
			# store actual directory and create destination folder (temporary or not)
			pwd = getcwd()
			if dest is None:
				dest = mkdtemp()
			else:
				runCommand(['mkdir', '-p', dest])

			# copy all the project (but exclude the dest/ in case it is in the same folder)
			runCommand(['rsync', '-a', join(self._path, '*'), dest, '--exclude', dest] + (['--exclude', compileErrorPath] if compileErrorPath else []))
			# copy all the useful files (promise.h, cadnaizer)
			runCommand(['cp', resource_filename(__name__, "extra/promise.h"), dest])
			runCommand(['cp', resource_filename(__name__, "extra/promise_header.h"), dest])
			runCommand(['cp', '-p', resource_filename(__name__, "extra/cadnaizer"), dest])

			# update the files that have to be changed
			for i, f in enumerate(self._files):
				f.createFile(self._types, dest, prefix='withoutcadna_' if cadna else '', promiseHeader=(i != 0))

			# move the  temp directory
			cd(dest)

			# cadnaize the project
			if cadna:
				for f in self._files:
					beginName, endName = split(f.fileName)
					filename = join(beginName, 'withoutcadna_' + endName)
					if not runCommand(['./cadnaizer', filename, '-o', f.fileName])[0]:
						cd(pwd)
						raise PromiseError("cadnaizer failed...")

			# compile it
			with Timing(self._compilations):
				for line in self._compile:
					res, msgError = runCommand(line.split(' '), errorsAsDegug=compilationErrorAsDebug, alias=self._alias)

					if not res:
						msg = "Compilation failed"
						if cadna:
							msg += "\nIs Cadna installed? Did you link your cade with Cadna with `-lcadnac`," \
							       " `-L$CADNA_PATH/lib` and `-I$CADNA_PATH/include` ?"
							if 'CADNA_PATH' not in environ:
								msg += '\nThe environment variable `CADNA_PATH` is not set, you should probably set it to Cadna path with'
								msg += '\n >>> export CADNA_PATH=/path/to/cadna'

						if compileErrorPath:
							# create compileError folder and copy all the files in it
							copyPath = join(pwd, compileErrorPath, 'attempt' + str(self.compilations[1]))
							runCommand(['mkdir', '-p', copyPath])
							runCommand(['cp', '-r', '.', copyPath])
							runFile = join(copyPath, "run.sh")
							# put all the compile commands in a `run.sh` file
							with open(runFile, 'w') as f:
								f.write("# Automatically generated by Promise (%s)\n\n" % datetime.now().strftime("%d/%m/%Y %H:%M:%S"))
								f.write('# Config=' + str(self.typesDict)+'\n\n')
								f.write("\n".join('#'+l for l in msgError))
								f.write("\n")
								f.write("\n".join(line for line in self._compile))
							#chmod(runFile, S_IXUSR)

						cd(pwd)
						raise PromiseCompilationError(msg)




			# run it and check if the execution has failed
			with Timing(self._executions):
				run = self._run.split(' ')
				ret, lines = runCommand(list(('./' if i == 0 else '') + r for i, r in enumerate(run)))
				if not ret:
					cd(pwd)
					raise PromiseError("Execution failed (should return 0)")

			# get the results in dictionary digits (and fill self._ref if cadna)
			if result and cadna:
				digits = {}
				self._ref = {}
				self._power = {}
				# parse the lines to get the variable name, 3 values and the nb of significant digits
				for li in lines:
					m = regDumpST.match(li)
					if m:
						varName, index, val1, val2, val3, nbDigits = m.groups()
						val = float.fromhex(val1)/3+float.fromhex(val2)/3+float.fromhex(val3)/3
						if varName not in digits:
							self._ref[varName] = []
							digits[varName] = []
						self._ref[varName].append(val)
						self._power[varName] = 10 ** (-self._nbDigits[1].get(varName, self._nbDigits[0]))
						digits[varName].append(int(nbDigits))

				# compare to expectations
				st = "\n".join("Variable %s: expect %d and found %d" % (var, self._nbDigits[1].get(var, self._nbDigits[0]), min(d))
				               for var, d in digits.items() if min(d) < self._nbDigits[1].get(var, self._nbDigits[0]))
				if st:
					cd(pwd)
					raise PromiseCompilationError(
						"The number of significant digits found with Cadna is lower than the expectation: ", st)

			elif result:
				# parse the result and store them in a dictionary
				values = {}
				for li in lines:
					m = regDump.match(li)
					if m:
						varName, index, val = m.groups()
						if varName not in values:
							values[varName] = []
						try:
							values[varName].append(float.fromhex(val))
						except ValueError as e:
							logger.error("ValueError val=%s\n%s" % (val, li))
							cd(pwd)
							raise ValueError from e
				# compare to the expectation
				try:
					for var, d in values.items():
						for v, ref in zip(d, self._ref[var]):
							if isnan(v) or isinf(v):
								raise StopIteration(var)
							if abs(ref-v) >= abs(ref) * self._power[var]:
								if not((ref == 0) and (abs(v) <= self._power[var])):
									raise StopIteration(var)
				except StopIteration as e:
					varFailed = e.value

			# go back to previous folder
			cd(pwd)

			# store the result (memoization)
			if not cadna:
				self._cache[frozenset(self._types.items())] = (varFailed is None) if result else None
		else:
			logger.debug('\U00002192 Nothing to do (using result in cache)')

		# log the result
		if result:
			logger.debug('result: ' + ("\U0001F44D" if varFailed is None else
			                           ("\U0001F44E (because of variable " + varFailed + ")")))

		return self._cache[frozenset(self._types.items())] if not cadna else (varFailed is None) if result else None


	def exportFinalResult(self, path):
		"""Produces the final files in the path folder"""
		runCommand(['mkdir', '-p', path])
		tempDirList = []
		for f in self._files:
			# check if the file path contains extra directory which is not already created
			tempMiddleDirList = f.fileName.split('/')
			if len(tempMiddleDirList) > 1:
				# it contains extra directory
				tempMiddleDir = join(*tempMiddleDirList[0:-1])
				if tempMiddleDir not in tempDirList:
					# the extra directory is not already created, so create it
					tempDirList.append(tempMiddleDir)
					runCommand(['mkdir', '-p', join(path, tempMiddleDir)])
			f.createFile(self._types, path, final=True)


	def exportParsedFiles(self, path):
		"""Produces the files after the parsing"""
		# change the types equal to their name
		for k in self._types:
			if isinstance(k, str):
				self._types[k] = '__PR_' + k + '__'
			else:
				self._types[k] = '__PROMISE' + str(k) + '__'
		# produce the files
		runCommand(['mkdir', '-p', path])
		tempDirList = []
		for f in self._files:
			tempMiddleDirList = f.fileName.split('/')
			if len(tempMiddleDirList) > 1:
				# it contains extra directory
				tempMiddleDir = join(*tempMiddleDirList[0:-1])
				if tempMiddleDir not in tempDirList:
					# the extra directory is not already created, so create it
					tempDirList.append(tempMiddleDir)
					runCommand(['mkdir', '-p', join(path, tempMiddleDir)])
			f.createFile(self._types, path, final=True, prefix='parsed_')


	def runDeltaDebug(self, lowest, highest, tempPath, doPause=False, compileErrorPath=None):
		"""Run the Delta/Debug with lowest/highest precision
		first, try with lowest and then run DD algorithm
		retur False if the deltaDebug is not successfull"""
		# status and progress bar
		n = len([v for v in self._types.values() if v == highest])
		nb = ceil(n*log2(n))
		with tqdm(desc="Delta-Debug iterations", position=1, total=nb, leave=False, ncols=80,
		          bar_format='%s{l_bar}%s{bar}%s| {n_fmt} it. [{rate_fmt}{postfix}]%s' %
		                     (Fore.LIGHTMAGENTA_EX, Fore.BLUE, Fore.LIGHTMAGENTA_EX, Fore.RESET)) as bar:
			with tqdm(total=0, position=0, bar_format='%s{desc}%s' % (Fore.LIGHTMAGENTA_EX, Fore.RESET), leave=False) as status:

				# copy the _types dict
				ty = dict(self._types)

				# try with the lowest precision (only for those that are equal to the highest),
				# and pass if there is an compilation error
				self.changeSomeTypes(lowest, [k for k, v in self._types.items() if v == highest])
				try:
					if self.compileAndRun(tempPath, compilationErrorAsDebug=True):
						logger.info("The format %s is enough for your expectation", lowest)
						return True
						
					if doPause:
						pause()

				except PromiseCompilationError as err:
					logger.debug('Result: \U0001F44E (' + str(err) + ')')


				# go back to previous _types dict and run the Delta-Debug algorithm
				self._types = dict(ty)
				dd = PromiseDD(self, lowest, highest, status, bar, tempPath, doPause, compileErrorPath)
				dd.run()

				# delete the status and progress bar
				status.set_description_str('')
				status.refresh()
				bar.bar_format = '{desc}'
				bar.set_description_str('')
				bar.refresh()

		# return False if anything has changed by DeltaDebug
		return len([v for v in self._types.values() if v == highest]) != n



from .promise import Promise 
import os

__version__ = '0.0.5'

curr_loc = os.path.dirname(os.path.realpath(__file__))

set_cadna_env = False

if not os.path.isfile(curr_loc+'/cadna/a.out'):
    if 'CADNA_PATH' in os.environ:
         set_cadna_env = False
    else:
        import logging
        logging.basicConfig()
        log = logging.getLogger(__file__)

        log.warning("Have not found CADNA path, please ensure CADNA is installed in this machine.")
        
else:
    set_cadna_env = True
    
if set_cadna_env:
    os.environ["CADNA_PATH"] = curr_loc+'/cadna/'



# coding: utf-8

import logging
from setuptools import setup
from distutils.errors import CCompilerError, DistutilsExecError, DistutilsPlatformError


def readme():
	"""import the readme file"""
	with open('README.rst') as f:
		return f.read()

__version__ = '0.0.5'
ext_errors = (CCompilerError, ModuleNotFoundError, DistutilsExecError, DistutilsPlatformError, IOError, SystemExit)

logging.basicConfig()
log = logging.getLogger(__file__)

setup_args = {'name':'cadnaPromise',
		'version':__version__,
		'description':'Precision auto-tuning of floating-point variables in program',
		'long_description':readme(),
		# packages = setuptools.find_packages(),
		'classifiers':["Intended Audience :: Science/Research",
					"Intended Audience :: Developers",
					"Programming Language :: C",
					"Programming Language :: Python",
					"Topic :: Software Development",
					"Topic :: Scientific/Engineering",
					'Operating System :: Microsoft :: Windows',
					'Operating System :: POSIX',
					'Operating System :: Unix',
					'Operating System :: MacOS',
					"Programming Language :: Python :: 3",
					],
		'keywords':'computer arithmetic mixed-precision, precision auto-tuning',
		'url':'https://github.com/PEQUAN/cadnaPromise',
		'author':'PEQUAN team',
		'author_email':'thibault.hilaire@lip6.fr; fabienne.jezequel@lip6.fr',
		'license':'GNU General Public License v3.0',
		'packages':{"cadnaPromise",
			  		"cadnaPromise.cadna", 
					"cadnaPromise.deltadebug",  
					"cadnaPromise.extra"
					},
		'package_data':{"cadnaPromise": ["deltadebug/*", "cadna/*", "extra/*"]},
		'tests_require':['pytest', 'pytest-cov'],
		'setup_requires':['colorlog', 'colorama', 'regex'],
		'install_requires':['colorlog', 'colorama', 'tqdm', 'regex', 'pytest', 'pytest-cov', 'yaml'],
		'extras_require':{'with_doc': ['sphinx', 'sphinx_bootstrap_theme']},
		'include_package_data':True,
		'data_files':[('extra/', ['cadnaPromise/extra/promise.h', 'cadnaPromise/extra/cadnaizer'])],
		'zip_safe':False,
		'entry_points':{'console_scripts': ['runPromise=cadnaPromise.run:runPromise',
										'promise=cadnaPromise.run:runPromise_custom', 
										'activate-promise=cadnaPromise.install:initCADNA',
										'test-promise=cadnaPromise.run:runPromise_custom_test']
				}}


try:	
	setup(**setup_args)

except ext_errors as ext_reason:
    log.warning("The installation is not successful, please contact the maintenance team.")



# coding=utf-8

"""This file is part of PROMISE.

	PROMISE is free software: you can redistribute it and/or modify it
	under the terms of the GNU Lesser General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	PROMISE is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
	Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with PROMISE. If not, see
	<http://www.gnu.org/licenses/>.

Promise v1 was written by Romain Picot
Promise v2 has been written from v1 by Thibault Hilaire and Sara Hoseininasab
	Sorbonne Université
	LIP6 (Computing Science Laboratory)
	Paris, France
Contact: thibault.hilaire@lip6.fr


File: logger.py
Date: April 2020

	Some useful functions for the logger
	Mainly, each file instanciates a PrLogger (that contains an instance of a logger)
	At the beginning, the logger is configured (from the yaml file and the command line options)

	Two handlers handle the logs:
	- one for display the information to the terminal
	- one for the log file
	The level set depends on the verbosity (--verbosity and --verbosityLog options). There are 5 levels of verbosity

	List of logging level, and the verbosity level (for each level, display everything above):
	- (40) ERROR (error messages)
	- (25) MESSAGE (start up, end messages)             <- verbosity 0
	- (20) INFO (steps done)                            <- verbosity 1
	- (10) DEBUG (some debug info)                      <- verbosity 2
	- (7) OUTPUT (output of the runs)
	- (6) COMMAND (command that are done)               <- verbosity 3
	- (5) LOW_DEBUG (low level debug messages)
	- (3) OUTPUT_PROMISE (values display by promise.h)  <- verbosity 4



	© Thibault HILAIRE, April 2020
"""


from colorlog import ColoredFormatter  # logging with colors
import logging
from os.path import join, expanduser
from .errors import PromiseError
from tqdm import tqdm


# logging level (see logging module) and their color, with respect to the verbosity
levels = [25, 20, 10, 6, 3]
log_colors = {
	'ERROR': 'red', 'MESSAGE': 'blue', 'INFO': 'white', 'OUTPUT': 'cyan,', 'DEBUG': 'green',
	'COMMAND': 'purple', 'LOW_DEBUG': 'yellow', 'OUTPUT_PROMISE': 'cyan'
}


class TqdmLoggingHandler(logging.Handler):
	"""	A dedicated Logging Handler
	to be used with tqdm progress bar, see
https://stackoverflow.com/questions/38543506/change-logging-print-function-to-tqdm-write-so-logging-doesnt-interfere-wit
	for more informationns"""
	def __init__(self, level=logging.NOTSET):
		"""Set up the super logger"""
		super().__init__(level)

	def emit(self, record):
		"""emit method, display message using tqdm.write instead of print"""
		msg = self.format(record)
		tqdm.write(msg)
		self.flush()



class PrLogger:
	"""Dedicated logger for Promise
	dedicated levels, dedicated handlers that are removed at the end"""

	def __init__(self):
		"""just get the root logger"""
		self._logger = logging.getLogger('promise')
		self._step = 'a'  # step

	def reset(self):
		"""Reset the logger"""
		self._step = 'a'
		for h in self._logger.handlers[:]:
			h.close()
			self._logger.removeHandler(h)

	# -- log levels --
	def error(self, msg, *args, **kws):
		"""log at message at level 40 (error)"""
		self._logger.log(40, msg, *args, **kws)

	def message(self, msg, *args, **kws):
		"""log at message at level 25 (message)"""
		self._logger.log(25, msg, *args, **kws)

	def info(self, msg, *args, **kws):
		"""log at message at level 20 (info)"""
		self._logger.log(20, msg, *args, **kws)

	def debug(self, msg, *args, **kws):
		"""log at message at level 10 (debug)"""
		self._logger.log(10, msg, *args, **kws)

	def output(self, msg, *args, **kws):
		"""log at message at level 5 (low level debug)"""
		self._logger.log(7, msg, *args, **kws)

	def command(self, msg, *args, **kws):
		"""log at message at level 5 (low level debug)"""
		self._logger.log(6, msg, *args, **kws)

	def lowdebug(self, msg, *args, **kws):
		"""log at message at level 5 (low level debug)"""
		self._logger.log(5, msg, *args, **kws)

	def outputPromise(self, msg, *args, **kws):
		"""log at message at level 5 (very low level debug)"""
		self._logger.log(3, msg, *args, **kws)

	def step(self, string):
		"""log a new step"""
		self._logger.info(self._step + ') ' + string)
		self._step = chr(ord(self._step) + 1)


	def configureLogger(self, options):
		"""Configure the logger from the options of the *command line* (only)
		(not from the configuration file, because we need to configure the logger
		before reading the configuration file, otherwise we cannot log the potential errors
		while parsing it)
		"""
		# get the path
		path = options['--path'] if options['--path'] else ''
		# get the verbosity
		verbosity = getInt(self._logger, options, 'verbosity', 1, 4, default=1)
		verbosityLog = getInt(self._logger, options, 'verbosityLog', 1, 4, default=verbosity)
		options['--verbosityLog'] = verbosityLog
		# set new logger levels for messages and low level debug
		logging.addLevelName(25, "MESSAGE")
		logging.addLevelName(7, "OUTPUT")
		logging.addLevelName(6, "COMMAND")
		logging.addLevelName(5, "LOW_DEBUG")
		logging.addLevelName(3, "OUTPUT_PROMISE")
		# stream the log to the console, with colors
		stream_handler = TqdmLoggingHandler()
		stream_handler.setLevel(levels[verbosity])
		stream_handler.setFormatter(ColoredFormatter("%(log_color)s%(message)s%(reset)s", log_colors=log_colors))
		self._logger.addHandler(stream_handler)
		# stream the log to the file
		if options['--log']:
			self.addFileLogger(path, options['--log'], verbosityLog)
		# set the logger to the minimum level (the handler will filter the messages)
		self._logger.setLevel(3)


	def addFileLogger(self, path, filename, verbosity):
		"""Add a file handler for the logger
		it is be called from `configureLogger` (when the command line options is parsed)
		and also by `getYMLoptions` when the yml file is parsed
		(the file logger should be configure *before* parsing the yml file,
		in order to log the potential errors of the parsing)"""
		file_handler = logging.FileHandler(expanduser(join(path, filename)), mode='w')
		file_handler.setLevel(levels[verbosity])
		error_formatter = logging.Formatter('%(asctime)s: %(message)s', "%m/%d %H:%M:%S")
		file_handler.setFormatter(error_formatter)
		self._logger.addHandler(file_handler)



def getInt(logger, options, name, mini, maxi, default=None):
	"""get an integer in the options, betwen min and max
	if not present, the default value is taken
	use the logger to log the errors, if happens"""
	optName = '--' + name
	try:
		val = int(options[optName]) if (optName in options) and (options[optName]) else default
	except ValueError as err:
		logger.error('The ' + name + ' is incorrect (`--' + name + '` option)')
		raise PromiseError(err) from err
	if not (mini <= val <= maxi):
		logger.error('The %s should be between %d and %d (included) (`--%s` option)' % (name, mini, maxi, name))
		raise PromiseError
	return val



